// Avago RGBC Sensor Library
// Code
//
// Code snipets used from Marcus @ Interactive Matter
// http://interactive-matter.org 
// https://wiki.engr.illinois.edu/display/ae498mpa/Color+Picker+--+Murph, +Joe, +Daniel
//
// Avago calibration procedure and videos from
// Lycee Monge de Chambery
// http://prepatsi.monge.free.fr/index.php?Itemid=71&id=65&option=com_content&task=view
//
// I2C writeTo() readFrom() functions
// Submitted by fabio on Sun, 2011-01-23 17:26.
// http://www.varesano.net/blog/fabio/help-me-pushing-writeto-and-readfrom-arduino-wire
// 
// http://sites.google.com/site/vilorei/
// Jan 05, 2011 - With I2C RGB LED with colour identification
// Jan 24, 2011 - Library


//#include <Wire.h>
#include <WProgram.h>
#include "I2C_RGBC_Reader.h"


// ---------------- Functions
// Writes value to address register on device
static void _writeTo(int device, byte register0, byte value) {
  Wire.beginTransmission(device); // start transmission to ACC 
  Wire.send(register0);       // send register address
  Wire.send(value);        // send value to write
  Wire.endTransmission();     // end transmission
}


// Reads number bytes starting from address register on device into buffer array
static void _readFrom(int device, byte register0, int number, byte buffer[]) {
  Wire.beginTransmission(device); //start transmission to ACC 
  Wire.send(register0);    //sends address to read from
  Wire.endTransmission(); //end transmission

    Wire.beginTransmission(device); //start transmission to ACC
  Wire.requestFrom(device, number);  // request bytes from ACC

  int i = 0;
  while(Wire.available())  //ACC may send less than requested (abnormal)
  { 
    buffer[i] = Wire.receive(); // receive a byte
    i++;
  }
  Wire.endTransmission(); //end transmission
}


static byte _readByteFrom(int device, byte register0) {
  Wire.beginTransmission(device);
  Wire.send(register0);
  Wire.endTransmission();

  Wire.requestFrom(device, 2);
  while (Wire.available()<1) { 
    ; 
  }
  int result = Wire.receive();
  return result;
}


static int _readIntFrom(int device, byte register0) {
  return _readByteFrom(device, register0) + (_readByteFrom(device, register0 + 1) << 8);
}


static void _performMeasurement(int device) {  
  _writeTo(device, 0x00, 0x01); // start sensing
  while(_readByteFrom(device, 0x00) != 0) { 
    ; 
  }   // waiting for a result
}



const byte CAP_RED = 0x06;
const byte CAP_GREEN = 0x07;
const byte CAP_BLUE = 0x08;
const byte CAP_CLEAR = 0x09;

const byte REG_INT_RED_LO = 0x0A;
const byte REG_INT_RED_HI = 0x0B;
const byte REG_INT_GREEN_LO = 0x0C;
const byte REG_INT_GREEN_HI = 0x0D;
const byte REG_INT_BLUE_LO = 0x0E;
const byte REG_INT_BLUE_HI = 0x0F;
const byte REG_INT_CLEAR_LO = 0x10;
const byte REG_INT_CLEAR_HI = 0x11;

const byte REG_DATA_RED_LO = 0x40;
const byte REG_DATA_RED_HI = 0x41;
const byte REG_DATA_GREEN_LO = 0x42;
const byte REG_DATA_GREEN_HI = 0x43;
const byte REG_DATA_BLUE_LO = 0x44;
const byte REG_DATA_BLUE_HI = 0x45;
const byte REG_DATA_CLEAR_LO = 0x46;
const byte REG_DATA_CLEAR_HI = 0x47;

const byte OFFSET_RED = 0x48;
const byte OFFSET_GREEN = 0x49;
const byte OFFSET_BLUE = 0x4A;
const byte OFFSET_CLEAR = 0x4B;
const int MAX_MEMORY_RGBC = 8;

static void _setGain(int device, int gainRegister, int gain) {
  if (gain < 4096) {
    byte hi = gain >> 8;
    byte lo = gain;

    _writeTo(device, gainRegister, lo);
    _writeTo(device, gainRegister+1, hi);
  }
}

// ----- Private -----


int I2C_RGBC_Reader::_calibrateClearGain() {        // 1- required white object
  int gainFound = 0;
  int upperBox=4096;
  int lowerBox = 0;
  int half;

  while (!gainFound) {
    half = ((upperBox-lowerBox)/2)+lowerBox;
    //no further halfing possbile
    if (half==lowerBox) {
      gainFound=1;
    }
    else {
      _setGain(_address, REG_INT_CLEAR_LO, half);
      _performMeasurement(_address);
      int halfValue = _readIntFrom(_address, REG_DATA_CLEAR_LO);

      if (halfValue>1000) {
        upperBox=half;
      }
      else if (halfValue<1000) {
        lowerBox=half;
      }
      else {
        gainFound=1;
      }
    }
  }
  return half;
}


int I2C_RGBC_Reader::_calibrateColorGain() {      // 2- required white object
  int gainFound = 0;
  int upperBox=4096;
  int lowerBox = 0;
  int half;

  while (!gainFound) {
    half = ((upperBox-lowerBox)/2)+lowerBox;
    //no further halfing possbile
    if (half==lowerBox) {
      gainFound=1;
    }
    else {
      _setGain(_address, REG_INT_RED_LO, half);
      _setGain(_address, REG_INT_GREEN_LO, half);
      _setGain(_address, REG_INT_BLUE_LO, half);

      _performMeasurement(_address);
      int halfValue = 0;

      halfValue=max(halfValue, _readIntFrom(_address, REG_DATA_RED_LO));
      halfValue=max(halfValue, _readIntFrom(_address, REG_DATA_GREEN_LO));
      halfValue=max(halfValue, _readIntFrom(_address, REG_DATA_BLUE_LO));

      if (halfValue>1000) {
        upperBox=half;
      }
      else if (halfValue<1000) {
        lowerBox=half;
      }
      else {
        gainFound=1;
      }
    }
  }
  return half;
}


void I2C_RGBC_Reader::_calibrateCapacitors() {        // 3- required white object
  int  calibrationRed = 0;
  int  calibrationBlue = 0;
  int  calibrationGreen = 0;
  int calibrated = 0;

  //need to store detect better calibration
  int oldDiff = 5000;

  while (!calibrated) {
    // calibrate the sensor

    // sensor gain setting (Avago app note 5330)
    // CAPs are 4bit (higher value will result in lower output)
    _writeTo(_address, CAP_RED, calibrationRed);
    _writeTo(_address, CAP_GREEN, calibrationGreen);
    _writeTo(_address, CAP_BLUE, calibrationBlue);

    // int colorGain = _calibrateColorGain();
    int colorGain = _readIntFrom(_address, REG_INT_RED_LO);
    _setGain(_address, REG_INT_RED_LO, colorGain);
    _setGain(_address, REG_INT_GREEN_LO, colorGain);
    _setGain(_address, REG_INT_BLUE_LO, colorGain);

    int maxRead = 0;
    int minRead = 4096;
    int red   = 0;
    int green = 0;
    int blue  = 0;
    for (int i=0; i<4 ;i ++) {
      _performMeasurement(_address);
      red   += _readIntFrom(_address, REG_DATA_RED_LO);
      green += _readIntFrom(_address, REG_DATA_GREEN_LO);
      blue  += _readIntFrom(_address, REG_DATA_BLUE_LO);
    }
    red   /= 4;
    green /= 4;
    blue  /= 4;

    maxRead = max(maxRead, red);
    maxRead = max(maxRead, green);
    maxRead = max(maxRead, blue);

    minRead = min(minRead, red);
    minRead = min(minRead, green);
    minRead = min(minRead, blue);

    int diff = maxRead - minRead;

    if (oldDiff != diff) {
      if ((maxRead==red) &&      (calibrationRed<15)) {
        calibrationRed++;
      }
      else if ((maxRead == green) && (calibrationGreen<15)) {
        calibrationGreen++;
      }
      else if ((maxRead == blue) && (calibrationBlue<15)) {
        calibrationBlue++;
      }
    }
    else {
      calibrated = 1;
    }
    oldDiff=diff;

    int rCal = calibrationRed;
    int gCal = calibrationGreen;
    int bCal = calibrationBlue;
  }

}





// ----- Public -----

I2C_RGBC_Reader::I2C_RGBC_Reader()
{
  _none=1;
  _address=0x74;   // 7 bits
}


String I2C_RGBC_Reader::WhoAmI() {
  String s="";
  if (_address < 0x10) s="0";
  s = s + String(_address, 0x10) +"h ADJD-S371 RGBC (sensor)";
  s = s + "";
  return s ;
}



void I2C_RGBC_Reader::begin() {
  for (int i=0; i<MAX_MEMORY_RGBC; i++) {
    _r[i]=0;
    _g[i]=0;
    _b[i]=0;
    _c[i]=0;  
  }
  _referenceDistance=0;

  _red=0;
  _green=0;
  _blue=0;
  _clear=0;

  _redFactor=1;
  _blueFactor=1;
  _greenFactor=1;
}


void I2C_RGBC_Reader::get() {
  _performMeasurement(_address);

  _red = _readIntFrom(_address, REG_DATA_RED_LO);
  _green = _readIntFrom(_address, REG_DATA_GREEN_LO);
  _blue = _readIntFrom(_address, REG_DATA_BLUE_LO);
  _clear = _readIntFrom(_address, REG_DATA_CLEAR_LO);
}


void I2C_RGBC_Reader::getRGBC(int &r, int &g, int &b, int &c) {
  get();

  r=_red;
  g=_green;
  b=_blue;
  c=_clear;
}


void I2C_RGBC_Reader::standard() {
  // 0- Initial values
  _writeTo(_address, CAP_RED,  0x09); // 0x05
  _writeTo(_address, CAP_GREEN, 0x09); // 0x05
  _writeTo(_address, CAP_BLUE, 0x02); // 0x01 
  _writeTo(_address, CAP_CLEAR, 0x05); // 0x01

  _writeTo(_address, REG_INT_RED_LO,  0xC4);
  _writeTo(_address, REG_INT_RED_HI,  0x09);
  _writeTo(_address, REG_INT_GREEN_LO, 0xC4);
  _writeTo(_address, REG_INT_GREEN_HI, 0x09);
  _writeTo(_address, REG_INT_BLUE_LO, 0xC4);
  _writeTo(_address, REG_INT_BLUE_HI, 0x09);
  _writeTo(_address, REG_INT_CLEAR_LO, 0xC4);
  _writeTo(_address, REG_INT_CLEAR_HI, 0x09);

  _writeTo(_address, OFFSET_RED,  0x7F);
  _writeTo(_address, OFFSET_GREEN, 0x7F);
  _writeTo(_address, OFFSET_BLUE, 0x7F);
  _writeTo(_address, OFFSET_CLEAR, 0x7F);

  _redFactor=1;
  _blueFactor=1;
  _greenFactor=1;
}

void I2C_RGBC_Reader::calibrate() {
  _calibrateClearGain();         // 1- required white object
  _calibrateColorGain();       // 2- required white object
  _calibrateCapacitors();         // 3- required white object
}



void I2C_RGBC_Reader::offset() {          // 4- no object and taped
  _writeTo(_address, 0x00, 0x02); // offsets acquisition

  while(_readByteFrom(_address, 0x00) != 0) {   // waiting for a result
  }
  _writeTo(_address, 0x01, 0x01); // offsets triming
  delay(100);
}


void I2C_RGBC_Reader::white() {    // 5- required: white object
  get();

  int m=2000; //bigger anyway
  m=min(m, _red);
  m=min(m, _green);
  m=min(m, _blue);

  _redFactor=((float)m*255.0)/(1000*(float)_red);
  _greenFactor=((float)m*255.0)/(1000*(float)_green);
  _blueFactor=((float)m*255.0)/(1000*(float)_blue);
}


void I2C_RGBC_Reader::memorise(int colourNumber) {
  if (colourNumber<MAX_MEMORY_RGBC) {
    _r[colourNumber]=0;
    _g[colourNumber]=0;
    _b[colourNumber]=0;
    _c[colourNumber]=0;

    for (int i=0; i<4; i++) {
      get();

      _r[colourNumber] += float(_red); 
      _g[colourNumber] += float(_green);
      _b[colourNumber] += float(_blue);
      _c[colourNumber] += float(_clear);

      delay(100);
    }

    _r[colourNumber] /= 4;
    _g[colourNumber] /= 4;
    _b[colourNumber] /= 4;
    _c[colourNumber] /= 4;

    _referenceDistance = 99999;
    for (int i=0; i<MAX_MEMORY_RGBC; i++) {
      for (int j=i; j<MAX_MEMORY_RGBC; j++) {
        if (j>i) {
          float d=sqrt( sq(float(_r[j])-_r[i]) + sq(float(_g[j])-_g[i]) + sq(float(_b[j])-_b[i]) + sq(float(_c[j])-_c[i]) );
          if ((d > 0) && (d < _referenceDistance)) _referenceDistance = d;
        }
      }
    }
  }
}


void I2C_RGBC_Reader::recognise(int &colourNumber, int &proximity) { 
  get();

  float d[MAX_MEMORY_RGBC];
  float minDistance = 2048;
  int n = 99;

  for (int i=0; i<=MAX_MEMORY_RGBC; i++) {
    d[i]=sqrt( sq(float(_red)-_r[i]) + sq(float(_green)-_g[i]) + sq(float(_blue)-_b[i]) + sq(float(_clear)-_c[i]) );
    if (d[i] < minDistance) {
      minDistance=d[i];
      n=i;
    }
  }

  colourNumber = n;
  proximity = int(100*minDistance/_referenceDistance); // distance gives an idea of the quality
}


void I2C_RGBC_Reader::parameters() {
  String s;

  _performMeasurement(_address);

  Serial.print("\tRED\tGREEN\tBLUE\tCLEAR\n");

  Serial.print("int\t");
  Serial.print(String(_readIntFrom(_address, REG_INT_RED_LO), HEX));
  Serial.print("\t");
  Serial.print(String(_readIntFrom(_address, REG_INT_GREEN_LO), HEX));
  Serial.print("\t");
  Serial.print(String(_readIntFrom(_address, REG_INT_BLUE_LO), HEX));
  Serial.print("\t");
  Serial.print(String(_readIntFrom(_address, REG_INT_CLEAR_LO), HEX));
  Serial.print("\n");

  Serial.print("cap\t");
  Serial.print(String(_readByteFrom(_address, CAP_RED), HEX));
  Serial.print("\t");
  Serial.print(String(_readByteFrom(_address, CAP_GREEN), HEX));
  Serial.print("\t");
  Serial.print(String(_readByteFrom(_address, CAP_BLUE), HEX));
  Serial.print("\t");
  Serial.print(String(_readByteFrom(_address, CAP_CLEAR), HEX));
  Serial.print("\n");

  Serial.print("offset\t");
  Serial.print(String(_readByteFrom(_address, OFFSET_RED), HEX));
  Serial.print("\t");
  Serial.print(String(_readByteFrom(_address, OFFSET_GREEN), HEX));
  Serial.print("\t");
  Serial.print(String(_readByteFrom(_address, OFFSET_BLUE), HEX));
  Serial.print("\t");
  Serial.print(String(_readByteFrom(_address, OFFSET_CLEAR), HEX));
  Serial.print("\n");

  Serial.print("white\t");
  Serial.print(String(int(_redFactor*100)));
  Serial.print("%\t");
  Serial.print(String(int(_greenFactor*100)));
  Serial.print("%\t");
  Serial.print(String(int(_blueFactor*100)));
  Serial.print("%\n");

  Serial.print("rgbc\t");
  Serial.print(String(_red, DEC));
  Serial.print("\t");
  Serial.print(String(_green, DEC));
  Serial.print("\t");
  Serial.print(String(_blue, DEC));
  Serial.print("\t");
  Serial.print(String(_clear, DEC));
  Serial.print("\n");

  Serial.print("\n\tRED\tGREEN\tBLUE\tCLEAR\n");

  for (int i=0;i<MAX_MEMORY_RGBC;i++) {
    Serial.print("#");
    Serial.print(String(i));
    Serial.print("\t");
    Serial.print(String(_r[i]));
    Serial.print("\t");
    Serial.print(String(_g[i]));
    Serial.print("\t");
    Serial.print(String(_b[i]));
    Serial.print("\t");
    Serial.print(String(_c[i]));
    Serial.print("\n");
    
    Serial.print("reference distance");
    Serial.print(_referenceDistance);
    Serial.print("\n");
  }
}